import copy
import logging
import astropy.visualization as viz
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from astropy.visualization import ZScaleInterval
from matplotlib.ticker import MaxNLocator
from mpl_toolkits.axes_grid1 import make_axes_locatable
from .observer import Observer
from .utils import numeric


class Plotter:
    """
    Takes an observer with WCS, Target, Plan, Instrument, Regions, and Corners and makes a finding chart.
    """

    def __init__(self, obs: Observer):
        self.obs = obs

    def plot(self):
        """not implemented yet"""
        pass

    def make_finding_chart(self, plot_refcat: bool = True, plot_simbad: bool = True,
                           savefig: bool = True, radius: numeric = 14) -> matplotlib.axes.Axes:
        """Make finding chart for a given observation."""
        obs = self.obs
        image = obs.get_image(radius=radius)
        zscale = ZScaleInterval()
        vmin, vmax = zscale.get_limits(image.data.flat)
        obs.wcs = obs.get_wcs(header=image.header)
        regions = obs.regions_to_physical()

        fig, ax = plt.subplots(figsize=(10, 10), subplot_kw={'projection': obs.wcs})

        for i, region in enumerate(regions):
            region.plot(ax=ax, edgecolor='cyan', linestyle='-.', label=obs.ins.region_names[i])

        self.plot_image(image.data, ax=ax, cmap='viridis', scale='linear', vmin=vmin, vmax=vmax)
        tar_pix = obs.wcs.all_world2pix(obs.target.ra, obs.target.dec, 0)
        ax.scatter(tar_pix[0], tar_pix[1], marker='*', s=250, label='SN', color='orange')

        if plot_refcat:
            refcat_stars = obs.refcat_coords.to_pixel(obs.wcs)
            ax.scatter(refcat_stars[0], refcat_stars[1], facecolors='none', edgecolors='red', zorder=4, alpha=0.8,
                       marker='o', s=100, label='Refcat2')
        if plot_simbad:
            simbad_stars = obs.simbad_coords.to_pixel(obs.wcs)
            ax.scatter(simbad_stars[0], simbad_stars[1], facecolors='none', edgecolors='orange', zorder=5, alpha=0.8,
                       marker='o', s=100, label='Simbad')

        ax.legend(fontsize=15)
        ax.set_title(f"{obs.target.info['target_name']} {obs.ins.__class__.__name__} FC")
        if savefig:
            fig.savefig(f"{obs.target.info['target_name']}_{obs.ins.__class__.__name__}_finding_chart.png")
        plt.show()
        return ax

    # Copied from flows image plotter.
    @staticmethod
    def plot_image(image, ax=None, scale='log', cmap=None, origin='lower', xlabel=None, ylabel=None, cbar=None,
                   clabel='Flux ($e^{-}s^{-1}$)', cbar_ticks=None, cbar_ticklabels=None, cbar_pad=None, cbar_size='4%',
                   title=None, percentile=95.0, vmin=None, vmax=None, offset_axes=None, color_bad='k', **kwargs):
        """
        Utility function to plot a 2D image.
        Parameters:
            image (2d array): Image data.
            ax (matplotlib.pyplot.axes, optional): Axes in which to plot.
                Default (None) is to use current active axes.
            scale (str or :py:class:`astropy.visualization.ImageNormalize` object, optional):
                Normalization used to stretch the colormap.
                Options: ``'linear'``, ``'sqrt'``, ``'log'``, ``'asinh'``, ``'histeq'``, ``'sinh'``
                and ``'squared'``.
                Can also be a :py:class:`astropy.visualization.ImageNormalize` object.
                Default is ``'log'``.
            origin (str, optional): The origin of the coordinate system.
            xlabel (str, optional): Label for the x-axis.
            ylabel (str, optional): Label for the y-axis.
            cbar (string, optional): Location of color bar.
                Choises are ``'right'``, ``'left'``, ``'top'``, ``'bottom'``.
                Default is not to create colorbar.
            clabel (str, optional): Label for the color bar.
            cbar_size (float, optional): Fractional size of colorbar compared to axes. Default='4%'.
            cbar_pad (float, optional): Padding between axes and colorbar.
            title (str or None, optional): Title for the plot.
            percentile (float, optional): The fraction of pixels to keep in color-trim.
                If single float given, the same fraction of pixels is eliminated from both ends.
                If tuple of two floats is given, the two are used as the percentiles.
                Default=95.
            cmap (matplotlib colormap, optional): Colormap to use. Default is the ``Blues`` colormap.
            vmin (float, optional): Lower limit to use for colormap.
            vmax (float, optional): Upper limit to use for colormap.
            color_bad (str, optional): Color to apply to bad pixels (NaN). Default is black.
            kwargs (dict, optional): Keyword arguments to be passed to :py:func:`matplotlib.pyplot.imshow`.
        Returns:
            :py:class:`matplotlib.image.AxesImage`: Image from returned
                by :py:func:`matplotlib.pyplot.imshow`.
        .. codeauthor:: Rasmus Handberg <rasmush@phys.au.dk>
        """

        logger = logging.getLogger(__name__)

        # Backward compatible settings:
        make_cbar = kwargs.pop('make_cbar', None)
        if make_cbar:
            raise FutureWarning("'make_cbar' is deprecated. Use 'cbar' instead.")
            if not cbar:
                cbar = make_cbar

        # Special treatment for boolean arrays:
        if isinstance(image, np.ndarray) and image.dtype == 'bool':
            if vmin is None: vmin = 0
            if vmax is None: vmax = 1
            if cbar_ticks is None: cbar_ticks = [0, 1]
            if cbar_ticklabels is None: cbar_ticklabels = ['False', 'True']

        # Calculate limits of color scaling:
        interval = None
        if vmin is None or vmax is None:
            if np.isnan(image).all():
                logger.warning("Image is all NaN")
                vmin = 0
                vmax = 1
                if cbar_ticks is None:
                    cbar_ticks = []
                if cbar_ticklabels is None:
                    cbar_ticklabels = []
            elif isinstance(percentile, (list, tuple, np.ndarray)):
                interval = viz.AsymmetricPercentileInterval(percentile[0], percentile[1])
            else:
                interval = viz.PercentileInterval(percentile)

        # Create ImageNormalize object with extracted limits:
        if scale in ('log', 'linear', 'sqrt', 'asinh', 'histeq', 'sinh', 'squared'):
            if scale == 'log':
                stretch = viz.LogStretch()
            elif scale == 'linear':
                stretch = viz.LinearStretch()
            elif scale == 'sqrt':
                stretch = viz.SqrtStretch()
            elif scale == 'asinh':
                stretch = viz.AsinhStretch()
            elif scale == 'histeq':
                stretch = viz.HistEqStretch(image[np.isfinite(image)])
            elif scale == 'sinh':
                stretch = viz.SinhStretch()
            elif scale == 'squared':
                stretch = viz.SquaredStretch()

            # Create ImageNormalize object. Very important to use clip=False here, otherwise
            # NaN points will not be plotted correctly.
            norm = viz.ImageNormalize(data=image, interval=interval, vmin=vmin, vmax=vmax, stretch=stretch, clip=False)

        elif isinstance(scale, (viz.ImageNormalize, matplotlib.colors.Normalize)):
            norm = scale
        else:
            raise ValueError("scale {} is not available.".format(scale))

        if offset_axes:
            extent = (offset_axes[0] - 0.5, offset_axes[0] + image.shape[1] - 0.5, offset_axes[1] - 0.5,
                      offset_axes[1] + image.shape[0] - 0.5)
        else:
            extent = (-0.5, image.shape[1] - 0.5, -0.5, image.shape[0] - 0.5)

        if ax is None:
            ax = plt.gca()

        # Set up the colormap to use. If a bad color is defined,
        # add it to the colormap:
        if cmap is None:
            cmap = copy.copy(plt.get_cmap('Blues'))
        elif isinstance(cmap, str):
            cmap = copy.copy(plt.get_cmap(cmap))

        if color_bad:
            cmap.set_bad(color_bad, 1.0)

        im = ax.imshow(image, cmap=cmap, norm=norm, origin=origin, extent=extent, interpolation='nearest', **kwargs)
        if xlabel is not None:
            ax.set_xlabel(xlabel)
        if ylabel is not None:
            ax.set_ylabel(ylabel)
        if title is not None:
            ax.set_title(title)
        ax.set_xlim([extent[0], extent[1]])
        ax.set_ylim([extent[2], extent[3]])

        if cbar:
            fig = ax.figure
            divider = make_axes_locatable(ax)
            if cbar == 'top':
                cbar_pad = 0.05 if cbar_pad is None else cbar_pad
                cax = divider.append_axes('top', size=cbar_size, pad=cbar_pad)
                orientation = 'horizontal'
            elif cbar == 'bottom':
                cbar_pad = 0.35 if cbar_pad is None else cbar_pad
                cax = divider.append_axes('bottom', size=cbar_size, pad=cbar_pad)
                orientation = 'horizontal'
            elif cbar == 'left':
                cbar_pad = 0.35 if cbar_pad is None else cbar_pad
                cax = divider.append_axes('left', size=cbar_size, pad=cbar_pad)
                orientation = 'vertical'
            else:
                cbar_pad = 0.05 if cbar_pad is None else cbar_pad
                cax = divider.append_axes('right', size=cbar_size, pad=cbar_pad)
                orientation = 'vertical'

            cb = fig.colorbar(im, cax=cax, orientation=orientation)

            if cbar == 'top':
                cax.xaxis.set_ticks_position('top')
                cax.xaxis.set_label_position('top')
            elif cbar == 'left':
                cax.yaxis.set_ticks_position('left')
                cax.yaxis.set_label_position('left')

            if clabel is not None:
                cb.set_label(clabel)
            if cbar_ticks is not None:
                cb.set_ticks(cbar_ticks)
            if cbar_ticklabels is not None:
                cb.set_ticklabels(cbar_ticklabels)

            # cax.yaxis.set_major_locator(matplotlib.ticker.AutoLocator())
            # cax.yaxis.set_minor_locator(matplotlib.ticker.AutoLocator())
            cax.tick_params(which='both', direction='out', pad=5)

        # Settings for ticks:
        integer_locator = MaxNLocator(nbins=10, integer=True)
        ax.xaxis.set_major_locator(integer_locator)
        ax.xaxis.set_minor_locator(integer_locator)
        ax.yaxis.set_major_locator(integer_locator)
        ax.yaxis.set_minor_locator(integer_locator)
        ax.tick_params(which='both', direction='out', pad=5)
        ax.xaxis.tick_bottom()
        ax.yaxis.tick_left()

        return im



import astropy.units as u
import numpy as np


class Corner:

    def __init__(self, x: u.Quantity, y: u.Quantity, hw: u.Quantity):
        self.x = x
        self.y = y
        self.hw = hw
        self.corner_xy = self.set_corners()
        self.corners = self.process_corner()

    def set_corners(self):
        """Get corners of a rectangle for a given ra dec and side length
        :return: list[corner 1, corner 2, ..]
        """
        return [(self.x - self.hw, self.y - self.hw), (self.x - self.hw, self.y + self.hw),
                (self.x + self.hw, self.y + self.hw), (self.x + self.hw, self.y - self.hw)]

    def process_corner(self):
        """Given corners of a rectangle defined as astropy Quantity objects, return it as an np array of floats"""
        _points = [u.quantity.Quantity(corner) for corner in self.corner_xy]
        return np.array(_points)



import pytest
import astropy.units as u
import tendrils
from .observer import get_flows_observer, Observer
from .auth import test_connection
from .plots import Plotter
from .utils import api_token

rot, tid, shifta, shiftd = 30, 8, 10, 10
token = api_token()


@pytest.mark.skip(reason="Skip until destructive overwrite is fixed")
def test_auth(monkeypatch):
    """
    Test the auth module.
    """
    with pytest.warns(RuntimeWarning):
        monkeypatch.setattr('builtins.input', lambda _: 'bad_token')
        test_connection()


@pytest.fixture
def observer():
    return get_flows_observer(rot, tid, shifta, shiftd)


def test_get_brightest(capsys, observer):
    observer.check_bright_stars(region=observer.regions[0])
    captured = capsys.readouterr()
    assert "Brightest star has" in captured.out


def test_plan(observer):
    assert observer.plan.rotation == rot * u.deg
    assert observer.plan.alpha == shifta * u.arcsec
    assert observer.plan.delta == shiftd * u.arcsec
    assert observer.plan.shift is True
    assert observer.plan.rotate is True


def test_observer(observer):
    isinstance(observer, Observer)


@pytest.mark.slow
def test_make_finding_chart(observer, monkeypatch):
    import matplotlib.pyplot as plt
    monkeypatch.setattr(plt, 'show', lambda: None)
    plotter = Plotter(observer)
    ax = plotter.make_finding_chart(observer, savefig=False)
    assert len(ax.get_images()) > 0
    title = ax.get_title()
    assert title.startswith(f"{observer.target.info['target_name']}")
    assert title.endswith("FC")



import os
import argparse
from typing import Union
import astropy.units as u
from astropy.coordinates import SkyCoord
from astropy.table import Table
import tendrils

# custom types
numeric = Union[float, int, u.Quantity]
tabular = tuple[SkyCoord, Table]


def parse():
    """Parse command line input to get target, position angle (rotate), alpha and delta offsets (shifta, shiftd)
    """
    parser = argparse.ArgumentParser(description='Calculate Brightest Star')
    parser.add_argument('-t', '--target', help="calculate for this targetname or targetid", type=str, default='None',
                        action='store')
    parser.add_argument('-r', '--rotate', help='rotation angle in degrees', type=float, default=0.0, action='store')
    parser.add_argument('-a', '--shifta', help='shift alpha in arcsec', type=float, default=0.0, action='store')
    parser.add_argument('-d', '--shiftd', help='shift delta in arcsec', type=float, default=0.0, action='store')
    parser.add_argument('-p', '--plot', help='whether to query images and plot', action='store_true')

    args = parser.parse_args()
    if args.target == 'None':
        parser.error('target id or name not provided, use -t <targetid> or <targetname>')
    elif args.target.isnumeric():
        args.target = int(args.target)
    return args.rotate, args.target, args.shifta, args.shiftd, args.plot


def api_token():
    """Try to get the API token from environment variable or from tendrils config.ini"""
    token = str(os.environ.get('FLOWS_API_TOKEN'))
    if token is not None and token != 'None':
        tendrils.utils.set_api_token(token=token, overwrite=True)
        return token

    cfg = tendrils.utils.load_config()
    token = cfg.get('api', 'token', fallback='None')
    if token.lower() in ['', 'none', 'bad', 'bad_token', 'badtoken']:
        token = None
    else:
        tendrils.utils.set_api_token(token=token, overwrite=True)
    return token


if __name__ == '__main__':
    token = api_token()
    tendrils.utils.set_api_token('None')



"""
Helper script for getting the brightest star in a Hawk-I (or other)
instrument Field of View for a given FLOWS target, retrieved via
tendrils API. Extensible for other instruments and projects.
Can optionally make a finding chart using NASA SkyView.
"""
from .version import __version__
from .run_get_brightest import main


from __future__ import annotations
import warnings
from erfa import ErfaWarning
from .plots import Plotter
from .auth import test_connection
from .observer import get_flows_observer
from .utils import parse

# Most useless warnings ever spammed for every operation by this package!
warnings.filterwarnings('ignore', category=ErfaWarning, append=True)


def main():
    # Parse input
    rot, tid, shifta, shiftd, make_fc = parse()

    # Test connection to flows:
    test_connection()

    # Print brightest star in field
    obs = get_flows_observer(rot, tid, shifta, shiftd)
    obs.check_bright_stars(region=obs.regions[0])

    # Make finding chart if requested
    if make_fc:
        Plotter(obs).make_finding_chart(radius=14)


if __name__ == '__main__':
    main()



from dataclasses import dataclass, field
from typing import Optional
from .utils import numeric
from astropy.coordinates import SkyCoord
import astropy.units as u


@dataclass
class Target:
    tid: int | str  # target id
    ra: numeric
    dec: numeric
    coords: Optional[SkyCoord] = None
    info: dict[str, ...] = field(default_factory=dict)

    def __post_init__(self):
        if self.coords is None:
            self.coords = SkyCoord(self.ra << u.deg, self.dec << u.deg, frame='icrs')



from dataclasses import dataclass
import numpy as np
import astropy.units as u
from .utils import numeric

@dataclass
class Plan:
    rotation: numeric = 0.0
    alpha: numeric = 0.0
    delta: numeric = 0.0
    rotate: bool = False
    shift: bool = False

    def __post_init__(self):
        self.shift = self.set_shift()
        self.rotate = self.set_rotation()
        self._sanitize_quanity_input()

    def set_rotation(self) -> bool:
        if self.rotation == 0.0:
            return False
        elif self.rotation != 0.0:
            return True

    def set_shift(self) -> bool:
        shift = (np.array((self.alpha, self.delta)) == 0.0).all()  # skip shift if alpha and delta 0
        return not shift

    def _sanitize_quanity_input(self):
        self.alpha = self.alpha << u.arcsec
        self.delta = self.delta << u.arcsec
        self.rotation = self.rotation << u.deg


import astropy.io.fits as fits
# noinspection PyProtectedMember
from astropy.io.fits.hdu.image import _ImageBaseHDU
from astropy.nddata import Cutout2D
from astropy.wcs import WCS
import matplotlib.pyplot as plt
from astropy.visualization import ZScaleInterval
from pathlib import Path
import argparse
from astropy.visualization.interval import BaseInterval
import numpy as np
from abc import ABC, abstractmethod

"""Expects fits compatible images saved with a single file extension (no .fits.gz), 
Makes a cutout based on values of y and x and recomputes wcs,
Saves the new fits file in output dir, 
preserves all extensions and applies the same cut to all image extensions
optionally plots the primary image in cut and uncut form.

code_author: emirkmo@github.com.
"""


class AbstractImage(ABC):

    @abstractmethod
    def __init__(self, image: _ImageBaseHDU):
        self.image = image
        self.ImType = type(image)
        self.wcs_init = None
        self.cut_image = None

    @abstractmethod
    def cutout(self, y: int, x: int):
        pass

    @abstractmethod
    def make_cutout(self, y: int, x: int):
        pass


# noinspection PyAttributeOutsideInit
class Image(AbstractImage):

    def __init__(self, image: _ImageBaseHDU,
                 primary: bool = False):
        self.image = image
        self.data = image.data
        self.header = image.header
        self.wcs_init = WCS(self.image.header)
        self.isprimary = primary
        self.ImType = type(image)
        if self.isprimary:
            self.wcs = self.wcs_init

    # def get_class(self):
    #    if isinstance(self.image,fits.ImageHDU):
    #        self.ImType = fits.ImageHDU
    #        self.isprimary = False
    #    elif isinstance(self.image,fits.PrimaryHDU):
    #        self.ImType = fits.PrimaryHDU
    #        self.isprimary = True
    #    else: raise(TypeError)

    def cutout(self, y, x):
        self.cut = Cutout2D(
            self.data,
            position=self.wcs.wcs.crpix,
            wcs=self.wcs,
            size=(y, x)
        )

    # noinspection PyCallingNonCallable
    def make_fits_from_cutout(self):
        self.cut_image = self.ImType(self.cut.data, self.cut.wcs.to_header())

    def find_missing_keys(self):
        self.missing_keys = set(self.header.keys()) - set(self.cut_image.header.keys())

    def fill_cut_image_header(self):
        for key in self.missing_keys:
            try:
                self.cut_image.header[key] = self.header[key]
            except ValueError:
                print(f'Bad fits key: {key}')

    def make_cutout(self, y: int, x: int):
        """convenience function for making a cutout image and forward filling header"""
        self.cutout(y, x)
        self.make_fits_from_cutout()
        self.find_missing_keys()
        self.fill_cut_image_header()


class Plotter:

    def __init__(self,
                 figsize: tuple = (8, 8),
                 zs: BaseInterval = ZScaleInterval(1000),
                 subplots: int = 1,
                 dpi: int = 125
                 ):
        self.figsize = figsize
        self.interval_finder = zs
        self.fig = plt.figure(figsize=self.figsize, dpi=dpi)
        self.subplots = subplots
        self.active_axis = 1
        self.axes = []
        self.ax = None

    # noinspection SpellCheckingInspection
    def plot_img(self, image: Image, wcs: WCS = 'None'):
        if wcs == 'None':
            wcs = image.wcs
        self.ax = self.fig.add_subplot(1, self.subplots, self.active_axis, projection=wcs)
        vmin, vmax = self.interval_finder.get_limits(image.data)
        self.ax.imshow(image.data, vmin=vmin, vmax=vmax)
        self.ax.grid()
        self.active_axis += 1
        self.axes.append(self.ax)


def parse():
    """Parse input
    :rtype: object
    """
    parser = argparse.ArgumentParser(prog='cut_image',
                                     description='Cut image and all image extensions')

    parser.add_argument("file", help="Specify the original file here")

    parser.add_argument("--output_dir", help="Give output directory for saving", default='./')
    parser.add_argument("-x", help='output suffix', type=int, default=4096)
    parser.add_argument("-y", help='output suffix', type=int, default=4030)
    parser.add_argument("-o", "--overwrite", help='overwrite output if exists',
                        action='store_true', default=False)
    parser.add_argument("-p", "--plot", help='overwrite output if exists',
                        action='store_true', default=False)

    return parser.parse_args()


def find_primary(hdul):
    # Find primary HDU and use its wcs
    primary_index = np.argwhere([isinstance(i, fits.PrimaryHDU) for i in hdul]).flatten()
    if primary_index.size <= 0 or primary_index.size > 1:
        raise ValueError('WCS to use could not be set as there were less than or greater than one Primary HDUs \
                         This should never be the case, please check your fits image!')
    return primary_index[0]


def get_primary(hdul) -> AbstractImage:
    return Image(hdul[find_primary(hdul)], primary=True)


def make_img_cutout(i, wcs, y, x) -> AbstractImage:
    img = Image(i, primary=False)
    img.wcs = wcs
    img.make_cutout(y, x)
    return img


def main():
    args = parse()
    curr = Path(args.file)
    save_path = Path.joinpath(Path(args.output_dir), curr.stem + '_cutout' + curr.suffix).resolve()

    hdul = fits.open(curr)
    new_hdul = fits.HDUList()

    primary_image = get_primary(hdul)
    primary_image.make_cutout(args.y, args.x)
    new_hdul.append(primary_image.cut_image)

    for i in hdul:
        if isinstance(i, fits.PrimaryHDU):
            continue
        elif isinstance(i, fits.ImageHDU):
            img = make_img_cutout(i, primary_image.wcs_init, args.y, args.x)
            new_hdul.append(img.cut_image)
        else:  # Mostly expecting extension to be a BinaryFitsTable.
            new_hdul.append(i)

    if args.plot:
        plotter = Plotter(subplots=2, dpi=200)
        plotter.plot_img(Image(new_hdul[0], primary=True))
        plotter.ax.set_title('cutout')
        plotter.plot_img(Image(hdul[0], primary=True))
        plotter.ax.set_title('uncut')
        plt.show(block=True)

    # Save
    new_hdul.writeto(str(save_path), output_verify='exception',
                     overwrite=args.overwrite, checksum=False)
    print(f'saved to: {save_path}')


if __name__ == '__main__':
    main()



import warnings
from astropy.coordinates import SkyCoord, Angle
from astroquery.simbad import Simbad
import astropy.units as u
from astroquery.skyview import SkyView
from astropy.table import Table
from typing import Optional


def query_simbad(coo_centre, radius=24*u.arcmin) -> tuple[Optional[Table], Optional[SkyCoord]]:
    """
    Query SIMBAD using cone-search around the position using astroquery.
    Parameters:
        coo_centre (:class:`astropy.coordinates.SkyCoord`): Coordinates of centre of search cone.
        radius (float, optional):
    Returns:
        list: Astropy Table with SIMBAD information.
    .. codeauthor:: Rasmus Handberg <rasmush@phys.au.dk>
    .. codeauthor:: Emir Karamehmetoglu <emir.k@phys.au.dk>
    """

    s = Simbad()
    s.ROW_LIMIT = 0
    s.remove_votable_fields('coordinates')
    s.add_votable_fields('ra(d;A;ICRS;J2000)', 'dec(d;D;ICRS;2000)', 'pmra', 'pmdec')
    s.add_votable_fields('otype')
    s.add_votable_fields('flux(B)', 'flux(V)', 'flux(R)', 'flux(I)', 'flux(J)', 'flux(H)', 'flux(K)')
    s.add_votable_fields('flux(u)', 'flux(g)', 'flux(r)', 'flux(i)', 'flux(z)')

    rad = Angle(radius).arcmin
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=UserWarning)
        results = s.query_criteria((f'region(circle, icrs, {coo_centre.icrs.ra.deg:.10f} '
                                    f'{coo_centre.icrs.dec.deg:+.10f}, {rad}m)'), otypes='Star')

    if not results:
        return None, None

    # Rename columns:
    results.rename_column('MAIN_ID', 'main_id')
    results.rename_column('RA_d_A_ICRS_J2000', 'ra')
    results.rename_column('DEC_d_D_ICRS_2000', 'dec')
    results.rename_column('PMRA', 'pmra')
    results.rename_column('PMDEC', 'pmdec')
    results.rename_column('FLUX_B', 'B_mag')
    results.rename_column('FLUX_V', 'V_mag')
    results.rename_column('FLUX_R', 'R_mag')
    results.rename_column('FLUX_I', 'I_mag')
    results.rename_column('FLUX_J', 'J_mag')
    results.rename_column('FLUX_H', 'H_mag')
    results.rename_column('FLUX_K', 'K_mag')
    results.rename_column('FLUX_u', 'u_mag')
    results.rename_column('FLUX_g', 'g_mag')
    results.rename_column('FLUX_r', 'r_mag')
    results.rename_column('FLUX_i', 'i_mag')
    results.rename_column('FLUX_z', 'z_mag')
    results.rename_column('OTYPE', 'otype')
    results.remove_column('SCRIPT_NUMBER_ID')
    results.sort(['V_mag', 'B_mag', 'H_mag'])

    # Filter out object types which shouldn'r really be in there anyway:
    indx = (results['otype'] == 'Galaxy') | (results['otype'] == 'LINER') | (results['otype'] == 'SN')
    results = results[~indx]

    if len(results) == 0:
        return None, None

    # Build sky coordinates object:
    simbad = SkyCoord(
        ra=results['ra'],
        dec=results['dec'],
        pm_ra_cosdec=results['pmra'],
        pm_dec=results['pmdec'],
        frame='icrs',
        obstime='J2000')

    return results, simbad


def query_2mass_image(ra, dec, pixels=2500, radius=50):
    radius = radius << u.arcmin
    out = SkyView.get_images(position='{}, {}'.format(ra, dec),
                             survey='2MASS-H', pixels=str(pixels),
                             coordinates='J2000', scaling='Linear', radius=radius)
    return out[0][0]



from abc import ABC, abstractmethod
from typing import Optional
from .target import Target
from .plan import Plan
from .corner import Corner
from .utils import numeric

from astropy.coordinates import SkyCoord
import astropy.units as u
import regions


class Instrument(ABC):
    """Instrument class"""

    @abstractmethod
    def __init__(self, coords: SkyCoord, rotation: Optional[u.Quantity] = 0.0 * u.deg):
        pass

    @abstractmethod
    def point(self, target: Target, plan: Plan) -> list[
        regions.RectangleSkyRegion, Optional[regions.RectangleSkyRegion]]:
        pass

    @abstractmethod
    def default_point(self):
        pass

    @abstractmethod
    def offset(self):
        pass

    @abstractmethod
    def get_regions(self):
        pass

    @property
    @abstractmethod
    def region_names(self):
        pass

    @abstractmethod
    def get_corners(self):
        pass


class Hawki(Instrument):
    """Class for storing the hardcoded chip and detector information of a VLT HAWKI pointing"""
    chip1_center = 3.75/2 * u.arcmin + 15 * u.arcsecond  # This is the distance to Chip 1 center from Field center.
    field_hw = 7.5 * u.arcminute  # full field height width
    chip1_hw = 3.5 * u.arcminute  # chip1 field height width
    chip1_offset = 112.5 * u.arcsecond  # default pointing offset of Hawki.

    def __init__(self, coords: SkyCoord, rotation: Optional[u.Quantity] = 0.0 * u.deg):
        # Coordinates
        self.rotation = rotation
        self.coords = coords  # initial (target) coord
        self.center_coords = self.default_point()  # Hawki center coord
        self.chip1_center_coords = self.offset()  # CCD4 center coord

        # Regions
        self.hawki_region, self.chip1_region = self.get_regions()

    def point(self, target: Target, plan: Plan) -> list[
        regions.RectangleSkyRegion, Optional[regions.RectangleSkyRegion]]:
        """point telescope to rot=rotation in degrees, alpha and delta offset in arcseconds"""
        self.coords = target.coords  # Assume unchanged
        if plan.rotate:
            self.rotation = plan.rotation

        self.center_coords = self.default_point(plan.alpha, plan.delta)
        self.chip1_center_coords = self.point_chip1(plan.alpha, plan.delta)
        self.hawki_region, self.chip1_region = self.get_regions()
        return [self.hawki_region, self.chip1_region]

    def offset(self, shifta: Optional[numeric] = 0.0, shiftd: Optional[numeric] = 0.0) -> SkyCoord:
        shifta = shifta << u.arcsecond
        shiftd = shiftd << u.arcsecond
        return self.coords.spherical_offsets_by(shifta, shiftd)

    def get_regions(self) -> tuple[regions.RectangleSkyRegion, regions.RectangleSkyRegion]:
        hawki_region = self.make_region(self.center_coords, self.field_hw, self.field_hw, self.rotation)
        chip1_region = self.make_region(self.chip1_center_coords, self.chip1_hw, self.chip1_hw, self.rotation)
        return hawki_region, chip1_region

    @property
    def region_names(self) -> tuple[str, str]:
        return 'HAWK-I', 'chip1'

    def get_corners(self) -> list[Corner, Corner]:
        chip1_corners = Corner(self.chip1_center_coords.ra, self.chip1_center_coords.dec, self.chip1_hw)
        field_corners = Corner(self.center_coords.ra, self.center_coords.dec, self.field_hw)
        return [field_corners, chip1_corners]

    @staticmethod
    def make_region(coords: SkyCoord, width: u.Quantity, height: u.Quantity, angle: u.Quantity):
        return regions.RectangleSkyRegion(coords, width=width, height=height, angle=angle)

    def default_point(self, alpha: u.Quantity = 0.0 * u.arcsec, delta: u.Quantity = 0.0 * u.arcsec) -> SkyCoord:
        return self.offset(-self.chip1_offset + alpha, -self.chip1_offset + delta)  # Hawki_center_coord

    def point_chip1(self, alpha: u.Quantity = 0.0 * u.arcsec, delta: u.Quantity = 0.0 * u.arcsec) -> SkyCoord:
        new_coords = self.chip1_center_coords.spherical_offsets_by(alpha + 15/2 * u.arcsecond, delta + 15/2 *
                                                                   u.arcsecond)
        """returns the CCD4 (chip1) ra and dec center given a pointing in ra dec and rotation"""
        return self.center_coords.directional_offset_by(
            self.center_coords.position_angle(new_coords) + self.rotation,
            self.center_coords.separation(new_coords))
        # return self.center_coords.spherical_offsets_by(self.chip1_center, self.chip1_center)

    @staticmethod
    def _get_pa_sep(coord1: SkyCoord, coord2: SkyCoord) -> tuple[u.Quantity, u.Quantity]:
        pa = coord1.position_angle(coord2)
        sep = coord1.separation(coord2)
        return pa, sep



import tendrils
from requests.exceptions import HTTPError
import warnings


def test_tendrils() -> bool:
    """
    Test Tendrils for API token authorization.
    """
    try:
        t = tendrils.api.get_target(8)
        return True
    except HTTPError as err:
        if 'Unauthorized' in err.response.reason:
            return False
        raise err  # This shouldn't happen. So raise it for debugging.


def update_api_token() -> None:
    token = input('Tendrils needs your FLOWS API token. Please enter it:')
    # We remove spaces from the token to avoid errors.
    tendrils.utils.set_api_token(token.strip().replace(' ', ''), overwrite=True)
    print('Token set. If there is a problem, please run `get_brightest` again.')


def test_connection() -> None:
    """
    Test connection to FLOWS API, and if faulty, prompt for a new token.
    """
    # Test connection to flows:
    if not test_tendrils():
        # Tendrils >0.3.0 will query first if token is None
        # But now we also query if connection fails after that too.
        # Useful when token is set to something invalid.
        update_api_token()
        warnings.warn(RuntimeWarning("Could not connect to flows via Tendrils. "
                                     "Check your API key and that target exists."
                                     "Also try Tendrils config steps from:"
                                     "https://www.github.com/SNFlows/tendrils/#before-you-begin-important"))



import argparse
import astropy.io.fits as fits
from astropy.wcs import WCS
import numpy as np
from pathlib import Path


def read_hdul(filepath, header_index=0):
    hdul = fits.open(filepath)
    wcs = WCS(hdul[header_index].header)
    return hdul, wcs


def swap_wcs(hdul_orig, hdul_diff):
    new_hdul = fits.HDUList()

    new_hdul.append(hdul_diff[0])
    new_hdul.append(hdul_orig[1])
    new_hdul.append(hdul_orig[2])
    return new_hdul


def get_basename_with_suffix(filename: Path, suffix=''):
    return Path(
        str(filename)[: str(filename).rfind(''.join(filename.suffixes))]
        + suffix
        + ''.join(filename.suffixes)
    )


def parse():
    parser = argparse.ArgumentParser(prog='FixTemplateSubHeaderforLCOGT', description='Attach header and extra dimensions from second \
                                     file to the first, while checking to make sure WCS is identical.\
                                     If not, uses WCS offirst file.')

    parser.add_argument("diff_file", help="Specify the diff file here")
    parser.add_argument("orig_file", help="Specify the original file here")
    parser.add_argument("--suffix", help='output suffix', type=str, default="_hfix")
    parser.add_argument("-o", "--overwrite", help='overwrite output if exists', action='store_true', default=False)

    return parser.parse_args()


def main():
    args = parse()

    diff_path = Path(args.diff_file)
    orig_path = Path(args.orig_file)

    # Get HDU list and WCS
    hdul_diff, wcs_diff = read_hdul(diff_path)
    hdul_orig, wcs_orig = read_hdul(orig_path)

    if wcs_diff.low_level_wcs == wcs_diff.low_level_wcs:
        save_file = get_basename_with_suffix(diff_path, args.suffix)
        new_hdul = swap_wcs(hdul_orig, hdul_diff)
        new_hdul.writeto(save_file, overwrite=args.overwrite)
        print(f'saved to: {save_file}')


if __name__ == '__main__':
    main()



from typing import Optional, Union

from astropy.coordinates import SkyCoord
import astropy.units as u
from astropy.time import Time
from astropy.wcs import WCS
from astropy.wcs.utils import celestial_frame_to_wcs
import regions
import numpy as np
from tendrils import api

from .instruments import Instrument, Hawki
from .target import Target
from .utils import numeric, tabular
from .catalogs import query_2mass_image, query_simbad
from .plan import Plan


class Observer:

    def __init__(self, instrument: Instrument, target: Target, plan: Plan, verbose: bool = False):
        self.target = target  # target info
        if not isinstance(self.target.coords, SkyCoord):
            raise TypeError('Target.coords must be an instance of SkyCoord representing target coordinates.')
        self.plan = plan  # Store offsets and rotations.
        self.ins = instrument  # store instrument specific info
        self.regions = self.ins.point(target, plan)
        self.nregions = len(self.regions)  # How many sub-regions of important passed by Instrument.?

        self.refcat_coords, self.refcat = self._make_refcat_catalog(mask=True, mask_dict={'H_mag': 14.0})
        self.simbad_coords, self.simbad = self._make_simbad_catalog()

        if not verbose:
            print((f'Simbad and refcat skyframes are '
                   f'equivalent: {self.refcat_coords.frame.is_equivalent_frame(self.simbad_coords.frame)}'
                   f'. If not, there might be a slight mismatch in alignment in current year.'))
        self.wcs = self.get_wcs(self.refcat_coords.frame)

    def _make_refcat_catalog(self, mask: bool = True, mask_dict: Optional[dict[str, float]] = None) -> tabular:
        refcat = api.get_catalog(self.target.tid)
        if mask and mask_dict is not None:
            masks = []
            keys_values = mask_dict.items()
            for key, value in keys_values:
                masks.append(refcat['references'][str(key)] <= float(value))
            if len(masks) == 1:
                refcat['references'] = refcat['references'][masks[0]]
            else:
                for msk in masks:
                    refcat['references'] = refcat['references'][msk]
        refcat_coords = SkyCoord(ra=refcat['references']['ra'], dec=refcat['references']['decl'],
                                 pm_ra_cosdec=refcat['references']['pm_ra'], pm_dec=refcat['references']['pm_dec'],
                                 obstime=Time(2015.5, format='decimalyear'))
        return refcat_coords, refcat['references']

    def _make_simbad_catalog(self) -> tabular:
        simbad = query_simbad(self.target.coords)
        # propagate Simbad catalog coords to refcat reference year
        simbad_coords = simbad[1].apply_space_motion(new_obstime=Time(2015.5, format='decimalyear'))
        return simbad_coords, simbad[0]

    def get_wcs(self, frame: Optional[object] = None, header: Optional[object] = None) -> WCS:
        if frame is not None:
            wcs = celestial_frame_to_wcs(frame)
            wcs.wcs.crval = np.array((self.target.ra, self.target.dec))
            return wcs
        elif header is not None:
            return WCS(header)
        else:
            raise AttributeError("Both frame and header were None, cannot calculate WCS")

    def get_image(self, pixels=2500, radius=50):
        return query_2mass_image(self.target.ra, self.target.dec, pixels, radius)

    def regions_to_physical(self) -> list[regions.RectangleSkyRegion]:
        pixel_regions = []
        for region in self.regions:
            pixel_regions.append(region.to_pixel(self.wcs))
        return pixel_regions

    def check_bright_stars(self, region: regions.RectangleSkyRegion, wcs: Optional[WCS] = None) -> np.ndarray:
        wcs = self.wcs if wcs is None else wcs
        # Check bright stars
        simbad_stars = region.contains(self.simbad_coords, wcs)
        catalog_stars = region.contains(self.refcat_coords, wcs)

        bright_stars = np.hstack((self.simbad[simbad_stars]['H_mag'].data, self.refcat[catalog_stars]['H_mag'].data))
        if np.ma.isMaskedArray(bright_stars):
            bright_stars = bright_stars.data[~np.isnan(bright_stars.data)]

        print('Brightest star has H-mag = {0:.1f}'.format(np.round(bright_stars.min(), 1)))
        return bright_stars


def get_flows_observer(rot: numeric, tid: Union[int, str], shifta: numeric, shiftd: numeric) -> Observer:
    """
    Returns the H-mag of the brightest star in the given region.
    """
    # Get query flows database
    target_info = api.get_target(tid)
    target_info['skycoord'] = SkyCoord(target_info['ra'] * u.deg, target_info['decl'] * u.deg)

    # Create Observer
    target = Target(tid, target_info['ra'], target_info['decl'], target_info['skycoord'], target_info)
    plan = Plan(rot, shifta, shiftd)
    hawki = Hawki(target.coords)
    return Observer(hawki, target, plan)
